@@ -51,11 +51,13 @@ def is_file(f):
     return os.path.isfile(os.path.expanduser(f))
 
 
-def zip_copy(dir_ins, dir_outs, filename, verbose=True, n_retry=1):
+def zip_copy(dir_ins, dir_outs, filename, filename_out=None, verbose=True, n_retry=1):
+    if not filename_out:
+        filename_out = filename
     for dir_in, dir_out in zip(dir_ins, dir_outs):
         if verbose:
-            print(f'Copy {dir_in}/{filename} -> {dir_out}/{filename}')
-        file_not_found_retry(shutil.copyfile, n_retry, f'{dir_in}/{filename}', f'{dir_out}/{filename}')
+            print(f'Copy {dir_in}/{filename} -> {dir_out}/{filename_out}')
+        file_not_found_retry(shutil.copyfile, n_retry, f'{dir_in}/{filename}', f'{dir_out}/{filename_out}')
 
 
 def zip_copytree(dir_ins, dir_outs, dirname, verbose=True, n_retry=1):
