@@ -12,22 +12,22 @@ add_cmd() {
 run_all() {
     if [ ! -z "${max_concurrent}" ]
     then
-        max_concurrent="-m {max_concurrent}"
+        max_concurrent="-m ${max_concurrent}"
     fi
     if [ ! -z "${env_source_file}" ]
     then
-        env_source_file="-e {env_source_file}"
+        env_source_file="-e ${env_source_file}"
     fi
     if [ ! -z "${nodes}" ]
     then
-        nodes="-n {nodes}"
+        nodes="-n ${nodes}"
     fi
 
     if [ -z ${1} ]
     then
         nodetool run_all ${nodes} ${max_concurrent} ${env_source_file} ${cmd_list_file}
     else
-        stdbuf -oL nodetool run_all ${nodes} ${max_concurrent} ${env_source_file} ${cmd_list_file} 2>&1 | tee ${1}
+        nodetool run_all --log_file ${1} ${nodes} ${max_concurrent} ${env_source_file} ${cmd_list_file}
     fi
 
     rm ${cmd_list_file}
