@@ -11,11 +11,14 @@ import getpass
 import asyncio
 import datetime
 import itertools
+import collections
+from enum import Enum
 
 import asyncssh
 import progressbar
 from click import style, secho
 
+
 localhost_shortname = socket.gethostname().split('.', 1)[0]
 
 
@@ -43,19 +46,39 @@ def expend_num_ranges(s):
 
 
 def get_hosts(host_string):
-    return list(set(host for k in host_string.split(',') for host in expend_num_ranges(k)))
+    hosts = list(set(host for k in host_string.split(',') for host in expend_num_ranges(k)))
+    return [localhost_shortname if h == 'localhost' else h for h in hosts]
 
 
 def get_worker_pool(name, nodes='localhost', max_concurrent=4, env_file=None, max_time=None,
                     debug=False, dry_run=False):
     hosts = get_hosts(nodes)
-    for i, host in enumerate(hosts):
-        if host == 'localhost':
-            hosts[i] = localhost_shortname
     return WorkerPool(hosts, name=name, max_tasks_per_worker=max_concurrent,
                       env_source_file=env_file, max_time=max_time, debug=debug, dry_run=dry_run)
 
 
+class TaskStatus(Enum):
+    WAITING = 1
+    RUNNING = 2
+    CANCELED = 3
+    SUCCESS = 4
+    ERROR = 5
+
+
+class WorkerStatus(Enum):
+    IDLE = 1
+    COMMITED = 2
+    RUNNING = 3
+    ERROR = 4
+
+
+class Process(object):
+
+    def __init__(self, client, process):
+        self.client = client
+        self.process = process
+
+
 class Task(object):
     """Represent a task executed by the worker pool
 
@@ -66,17 +89,19 @@ class Task(object):
         returncode (int): Return code. None if task was not executed.
     """
 
-    def __init__(self, name, command, output_file=None, done_callback=None):
+    def __init__(self, name, command, output_file=None, done_callback=None, run_on_host=None, n_try=3):
         self.name = name
         self.command = command
         self.output_file = output_file
         self.fd = None
         self.returncode = None
-        self.process = None
         self.n_try = 0
-        self.remote_host = None
-        self.remote_user = None
         self.done_callback = done_callback
+        self.run_on_host = run_on_host
+        self.status = TaskStatus.WAITING
+        self.process = None
+        self.n_try = n_try
+        self.timed_out = False
 
     def init_log(self):
         if self.output_file is not None:
@@ -84,20 +109,18 @@ class Task(object):
             self.fd.write(f'# Logging starting at {datetime.datetime.now()}\n')
             self.fd.write(f'# Input command: {self.command}\n\n')
 
-    def set_process(self, process, remote_host=None, remote_user=None):
+    def set_process(self, process):
         self.process = process
-        self.remote_host = remote_host
-        self.remote_user = remote_user
 
     async def terminate(self):
         if self.process is not None:
-            if hasattr(self.process, 'pid'):
-                if self.remote_host is None:
-                    self.log(f'Killing local process PID {self.process.pid}', 'local', err=True)
-                    os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
+            if hasattr(self.process.process, 'pid'):
+                if self.process.client.host is not 'local':
+                    self.log(f'Killing local process PID {self.process.process.pid}\n', 'local', err=True)
+                    os.killpg(os.getpgid(self.process.process.pid), signal.SIGTERM)
                 else:
-                    self.log(f'Killing remote process PID {self.process.pid}', self.remote_host, err=True)
-                    kill_remote(self.remote_host, self.remote_user, self.process.pid)
+                    self.log(f'Killing remote process PID {self.process.pid}\n', self.process.client.host, err=True)
+                    kill_remote(self.process.client.host, self.process.client.user, self.process.pid)
 
     def log(self, line, host, err=False):
         if isinstance(line, bytes):
@@ -110,8 +133,14 @@ class Task(object):
         print(f'[{self.name}:{host}] {line}')
 
     async def log_stream(self, stream, host, err=False):
-        async for line in stream:
-            self.log(line, host, err=err)
+        while True:
+            try:
+                async for line in stream:
+                    self.log(line, host, err=err)
+            except ValueError:
+                self.log('(line too long)', host, err=True)
+                continue
+            break
 
     def close_log(self):
         if self.fd is not None:
@@ -121,6 +150,30 @@ class Task(object):
             except Exception:
                 print(f'Error closing log file for task {self.name}. Ignoring.')
 
+    async def run(self, client_pool):
+        self.init_log()
+
+        for i in range(self.n_try):
+            self.timed_out = False
+
+            worker = await client_pool.get_worker(self.run_on_host)
+            self.status = TaskStatus.RUNNING
+
+            ret_status, err_msg = await worker.run(self)
+
+            if err_msg:
+                self.log(err_msg, worker.client.host, err=True)
+
+            if ret_status and not self.timed_out:
+                self.status = TaskStatus.SUCCESS
+                break
+
+            self.status = TaskStatus.ERROR
+            if i + 1 < self.n_try:
+                self.log(f'Will retry {i + 1} / {self.n_try} ...\n', worker.client.host, err=True)
+
+        self.close_log()
+
 
 class SSHClient(asyncssh.SSHClient):
 
@@ -140,9 +193,10 @@ class SSHClient(asyncssh.SSHClient):
 
 class Client(object):
 
-    def __init__(self, host, user=None, force_sync=True):
+    def __init__(self, host, user=None, force_sync=True, password=None):
         self.host = host
         self.user = user
+        self.password = password
         self.conn = None
         self.client = None
         self.starting = False
@@ -156,7 +210,8 @@ class Client(object):
             self.starting = True
             print(f'Starting client {self.host} ...')
             try:
-                self.conn, self.client = await asyncssh.create_connection(SSHClient, self.host, username=self.user)
+                self.conn, self.client = await asyncssh.create_connection(SSHClient, self.host, username=self.user,
+                                                                          password=self.password)
             except Exception:
                 raise
             finally:
@@ -168,7 +223,7 @@ class Client(object):
 
         async with self.creating_session:
             process = await self.conn.create_process('echo $$;' + task.command)
-        task.set_process(process, remote_host=self.host, remote_user=self.user)
+        task.set_process(Process(self, process))
 
         process.pid = await process.stdout.readline()
 
@@ -194,8 +249,8 @@ class Client(object):
 
 class LocalClient(object):
 
-    def __init__(self, force_sync=True):
-        self.host = 'local'
+    def __init__(self, host='local', force_sync=True):
+        self.host = host
         self.started = asyncio.Event()
         self.force_sync = force_sync
 
@@ -205,7 +260,7 @@ class LocalClient(object):
     async def execute(self, task):
         process = await asyncio.create_subprocess_shell(task.command, stdout=asyncio.subprocess.PIPE,
                                                         stderr=asyncio.subprocess.PIPE, preexec_fn=os.setpgrp)
-        task.set_process(process)
+        task.set_process(Process(self, process))
 
         await task.log_stream(process.stdout, self.host)
         await task.log_stream(process.stderr, self.host, err=True)
@@ -226,70 +281,58 @@ class LocalClient(object):
 
 class Worker(object):
 
-    def __init__(self, client, name, pool, max_try=3):
+    def __init__(self, client, name):
         self.client = client
         self.name = name
-        self.pool = pool
-        self.running = True
+        self.status = WorkerStatus.IDLE
         self.execute_start_time = None
         self.running_task = None
-        self.max_try = max_try
 
-    async def run(self):
+    async def run(self, task):
+        self.running_task = task
+        self.execute_start_time = time.time()
+        self.status = WorkerStatus.RUNNING
+        err_msg = ''
+
         try:
-            while True:
-                task = await self.pool.queue.get()
-
-                if task.n_try > 0:
-                    await asyncio.sleep(2)
-
-                try:
-                    task.init_log()
-                    await self.client.start()
-                    await self.client.started.wait()
-                    self.execute_start_time = time.time()
-                    self.running_task = task
-                    await self.client.execute(task)
-
-                    if task.done_callback is not None:
-                        try:
-                            task.done_callback()
-                        except Exception as exc:
-                            task.log(f'Error executing return callback of task {self.name}: {str(exc)}',
-                                     self.client.host, err=True)
-                except Exception as exc:
-                    task.n_try += 1
-                    if task.n_try < self.max_try:
-                        await asyncio.sleep(1)
-                        self.pool.queue.put_nowait(task)
-                        task.log(f'Error executing task: {str(exc)}, will retry ...', self.client.host, err=True)
-                    else:
-                        task.log(f'Error executing task {self.name}: {str(exc)}', self.client.host, err=True)
-                    self.pool.queue.task_done()
-                    self.running = False
-                    break
-                finally:
-                    self.execute_start_time = None
-                    self.running_task = None
-                    task.close_log()
-
-                self.pool.queue.task_done()
-                self.pool.tasks_done.append(task)
-                self.pool.pbar.update(len(self.pool.tasks_done))
-
-        except asyncio.CancelledError:
-            self.client.close()
+            await self.client.start()
         except Exception as exc:
-            secho(f'Error in worker {self.name}: {str(exc)}', fg='red')
-            raise
-        finally:
-            self.running = False
+            err_msg = f'Error starting client: {str(exc)}\n'
+            self.status = WorkerStatus.ERROR
+            return False, err_msg
+
+        await self.client.started.wait()
+
+        try:
+            await self.client.execute(task)
+        except ConnectionError as exc:
+            err_msg = f'Error connecting client: {str(exc)}\n'
+            self.status = WorkerStatus.ERROR
+            return False, err_msg
+        except Exception as exc:
+            err_msg = f'Error executing task {self.name}: {str(exc)}\n'
+            self.status = WorkerStatus.IDLE
+            return False, err_msg
+
+        if task.done_callback is not None:
+            try:
+                task.done_callback()
+            except Exception as exc:
+                err_msg = f'Error executing return callback of task {self.name}: {str(exc)}\n'
+
+        self.status = WorkerStatus.IDLE
+
+        return True, err_msg
+
+    def done(self):
+        self.status = WorkerStatus.IDLE
 
 
 class WorkerPool(object):
 
     def __init__(self, hosts, name='Worker', max_tasks_per_worker=4,
-                 env_source_file=None, user=None, max_time=None, force_sync=False, debug=False, dry_run=False):
+                 env_source_file=None, user=None, max_time=None, force_sync=False,
+                 debug=False, dry_run=False, password=None):
         """Initiate a worker pool.
 
         Args:
@@ -307,27 +350,50 @@ class WorkerPool(object):
 
         self.workers = []
         self.tasks = []
-        self.queue = asyncio.Queue()
-        self.tasks_done = []
-        self.tasks_error = []
-        self.tasks_timeout = []
         self.env_file = env_source_file
         self.max_time = max_time
-        self.pbar = None
         self.name = name
         self.debug = debug
         self.dry_run = dry_run
+        self.hosts = hosts
+        self.futures = None
+        self.clients = []
+        self.active_hosts = hosts
 
-        for host in hosts:
+        for host in self.hosts:
             if host == localhost_shortname:
-                client = LocalClient(force_sync=force_sync)
+                client = LocalClient(localhost_shortname, force_sync=force_sync)
             else:
-                client = Client(host, user, force_sync=force_sync)
+                client = Client(host, user, force_sync=force_sync, password=password)
+
             for i in range(max_tasks_per_worker):
-                worker = Worker(client, f'{client.host}:{i}', self)
+                worker = Worker(client, f'{client.host}:{i}')
                 self.workers.append(worker)
 
-    def add(self, command, name=None, output_file=None, done_callback=None):
+        widgets = [
+            f"{style(self.name, bold=True)}: ", progressbar.Percentage(), ' (',
+            progressbar.SimpleProgress(), ')'
+            ' ', progressbar.Bar(marker='|', left='[', right=']'),
+            ' ', progressbar.Timer(),
+            ' ', progressbar.ETA(),
+        ]
+
+        self.pbar = progressbar.ProgressBar(redirect_stdout=True, widgets=widgets)
+
+    async def get_worker(self, host=None):
+        while True:
+            for worker in self.workers:
+                if host is not None and worker.client.host != host:
+                    if host not in self.active_hosts:
+                        print(f'Requested run_on_host "{host}" not available anymore. Releasing constraint ...')
+                        host = None
+                    continue
+                if worker.status is WorkerStatus.IDLE:
+                    worker.status = WorkerStatus.COMMITED
+                    return worker
+            await asyncio.sleep(0.1)
+
+    def add(self, command, name=None, output_file=None, done_callback=None, run_on_host=None, n_try=3):
         """Add a command to execute by a worker in the pool. Optionally output result in output_file.
 
         Args:
@@ -335,6 +401,9 @@ class WorkerPool(object):
             name (str, optional): Name of the task. If not set it will be set by the worker pool
             output_file (str, optional): Optional filename to log output into.
         """
+        if run_on_host is not None and run_on_host not in self.hosts:
+            print(f'Requested run_on_host "{run_on_host}" is not in the hosts list. Releasing constraint ...')
+
         if self.env_file:
             command = "sh -c '. %s; %s'" % (self.env_file, command)
         if name is None:
@@ -345,92 +414,76 @@ class WorkerPool(object):
         if self.dry_run:
             return
 
-        self.tasks.append(Task(name, command, output_file=output_file, done_callback=done_callback))
+        self.tasks.append(Task(name, command, output_file=output_file, done_callback=done_callback,
+                               run_on_host=run_on_host, n_try=n_try))
 
-    async def _process_queue(self):
-        for task in self.tasks:
-            await self.queue.put(task)
-
-        self.pbar.start(max_value=self.queue.qsize())
-
-        futures = []
-        for worker in self.workers:
-            future = asyncio.ensure_future(worker.run())
-            worker.future = future
-            futures.append(future)
-        futures.append(asyncio.ensure_future(self._monitor_queue()))
-
-        # Wait until the queue is fully processed.
-        await self.queue.join()
-
-        # Cancel our worker tasks.
-        for future in futures[::-1]:
-            future.cancel()
-        # Wait until all worker tasks are cancelled.
-        await asyncio.gather(*futures, return_exceptions=False)
-
-        self.pbar.finish(dirty=len(self.tasks_error) > 0 or self.queue.qsize() > 0)
+    def get_all_tasks(self):
+        tasks = collections.defaultdict(list)
+        for t in self.tasks:
+            tasks[t.status].append(t)
+        return tasks
 
     async def _monitor_queue(self, update_interval=1):
         try:
-            while True:
+            n_todo = len(self.tasks)
+            while n_todo > 0:
                 await asyncio.sleep(update_interval)
-                self.pbar.update(len(self.tasks_done))
-                active_workers = 0
-                for w in self.workers:
-                    if w.running:
-                        active_workers += 1
-                        if w.execute_start_time is not None and self.max_time is not None:
-                            running_time = time.time() - w.execute_start_time
-                            if running_time > (self.max_time + update_interval):
-                                task = w.running_task
-                                task.log(f'Task timeout: waited for {running_time:.2f} s', w.client.host, err=True)
-                                task.log(f'Command was: {task.command}', w.client.host, err=True)
-                                if task.n_try < w.max_try - 1:
-                                    print(task.n_try, w.max_try)
-                                    task.log('Will retry ...', w.client.host, err=True)
-                                    task.n_try += 1
-                                    self.queue.put_nowait(task)
-                                    self.pbar.max_value += 1
-                                await task.terminate()
+                tasks = self.get_all_tasks()
+                n_done = len(tasks[TaskStatus.SUCCESS])
+                n_todo = len(tasks[TaskStatus.WAITING]) + len(tasks[TaskStatus.RUNNING])
+                self.pbar.update(n_done)
 
-                if active_workers == 0:
-                    print(f"No active workers, canceling all remaining tasks ...")
-                    print(f"{self.queue.qsize()}")
-                    for i in range(self.queue.qsize()):
-                        self.tasks_error.append(self.queue.get_nowait())
-                        self.queue.task_done()
-                    print(f"{self.queue.qsize()}")
+                self.active_hosts = set(w.client.host for w in self.workers if w.status != WorkerStatus.ERROR)
+                n_active_workers = 0
+
+                for worker in self.workers:
+                    n_active_workers += worker.status != WorkerStatus.ERROR
+
+                    if worker.status == WorkerStatus.RUNNING and self.max_time is not None:
+                        running_time = time.time() - worker.execute_start_time
+                        if running_time > (self.max_time + update_interval):
+                            task = worker.running_task
+                            task.timed_out = True
+                            task.log(f'Task timeout: waited for {running_time:.2f} s', worker.client.host, err=True)
+                            task.log(f'Command was: {task.command}', worker.client.host, err=True)
+                            await task.terminate()
+
+                if n_active_workers == 0:
+                    secho("No active workers, canceling all remaining tasks ...", fg='red')
+                    for future in self.futures[::-1]:
+                        future.cancel()
 
         except asyncio.CancelledError:
             pass
 
+    async def process_tasks(self):
+        self.pbar.start(max_value=len(self.tasks))
+
+        self.futures = [asyncio.ensure_future(task.run(self)) for task in self.tasks]
+        self.futures.append(asyncio.ensure_future(self._monitor_queue()))
+
+        await asyncio.gather(*self.futures, return_exceptions=False)
+
     def execute(self):
-        """Execute all queue tasks.
-
-        Returns:
-            (list_done, list_error): Return a list of tasks executed successfully and tasks not executed.
-        """
-        widgets = [
-            f"{style(self.name, bold=True)}: ", progressbar.Percentage(), ' (',
-            progressbar.SimpleProgress(), ')'
-            ' ', progressbar.Bar(marker='|', left='[', right=']'),
-            ' ', progressbar.Timer(),
-            ' ', progressbar.ETA(),
-        ]
-        self.pbar = progressbar.ProgressBar(redirect_stdout=True, widgets=widgets)
-
         if not self.tasks:
             self.pbar.finish()
             return [], []
 
         loop = asyncio.get_event_loop()
-        loop.run_until_complete(self._process_queue())
 
-        return self.tasks_done, self.tasks_error
+        try:
+            loop.run_until_complete(self.process_tasks())
+        except asyncio.CancelledError:
+            secho(f'Error: one (or more) task(s) have been canceled', fg='red')
+        finally:
+            for client in self.clients:
+                client.close()
 
-    def wait(self):
-        self.execute()
+        tasks = self.get_all_tasks()
+        n_success = len(tasks[TaskStatus.SUCCESS])
 
-    def close(self):
-        pass
+        self.pbar.update(n_success)
+        self.pbar.finish(dirty=n_success != len(self.tasks))
+
+        return tasks[TaskStatus.SUCCESS], tasks[TaskStatus.ERROR] + tasks[TaskStatus.CANCELED] \
+            + tasks[TaskStatus.WAITING] + tasks[TaskStatus.RUNNING]
